local Units = {unitFrames = {}, loadedUnits = {}, unitEvents = {}}
local stateMonitor = CreateFrame("Frame", nil, nil, "SecureHandlerBaseTemplate")
local unitFrames, loadedUnits, unitEvents, queuedCombat = Units.unitFrames, Units.loadedUnits, Units.unitEvents, {}
local playerClass = select(2, UnitClass("player"))
local _G = getfenv(0)

ShadowUF.Units = Units
ShadowUF:RegisterModule(Units, "units")

-- Frame shown, do a full update
local function FullUpdate(self)
	for i=1, #(self.fullUpdates), 2 do
		local handler = self.fullUpdates[i]
		handler[self.fullUpdates[i + 1]](handler, self)
	end
end

-- Register an event that should always call the frame
local function RegisterNormalEvent(self, event, handler, func)
	-- Make sure the handler/func exists
	if( not handler[func] ) then
		error(string.format("Invalid handler/function passed for %s on event %s, the function %s does not exist.", self:GetName() or tostring(self), event, func), 3)
		return
	end

	self:RegisterEvent(event)
	self.registeredEvents[event] = self.registeredEvents[event] or {}
	
	-- Each handler can only register an event once per a frame.
	if( self.registeredEvents[event][handler] ) then
		return
	end
			
	self.registeredEvents[event][handler] = func
end

-- Unregister an event
local function UnregisterEvent(self, event, handler)
	if( self.registeredEvents[event] ) then
		self.registeredEvents[event][handler] = nil
		
		local hasHandler
		for handler in pairs(self.registeredEvents[event]) do
			hasHandler = true
			break
		end
		
		if( not hasHandler ) then
			self:UnregisterEvent(event)
		end
	end
end

-- Register an event thats only called if it's for the actual unit
local function RegisterUnitEvent(self, event, handler, func)
	unitEvents[event] = true
	RegisterNormalEvent(self, event, handler, func)
end

-- Register a function to be called in an OnUpdate if it's an invalid unit (targettarget/etc)
local function RegisterUpdateFunc(self, handler, func)
	if( not handler[func] ) then
		error(string.format("Invalid handler/function passed to RegisterUpdateFunc for %s, the function %s does not exist.", self:GetName() or tostring(self), event, func), 3)
		return
	end

	for i=1, #(self.fullUpdates), 2 do
		local data = self.fullUpdates[i]
		if( data == handler and self.fullUpdates[i + 1] == func ) then
			return
		end
	end
	
	table.insert(self.fullUpdates, handler)
	table.insert(self.fullUpdates, func)
end

-- Used when something is disabled, removes all callbacks etc to it
local function UnregisterAll(self, handler)
	for i=#(self.fullUpdates), 1, -1 do
		if( self.fullUpdates[i] == handler ) then
			table.remove(self.fullUpdates, i + 1)
			table.remove(self.fullUpdates, i)
		end
	end

	for event, list in pairs(self.registeredEvents) do
		list[handler] = nil
		
		local hasRegister
		for handler in pairs(list) do
			hasRegister = true
			break
		end
		
		if( not hasRegister ) then
			self:UnregisterEvent(event)
		end
	end
end

-- Event handling
local function OnEvent(self, event, unit, ...)
	if( not unitEvents[event] or self.unit == unit ) then
		for handler, func in pairs(self.registeredEvents[event]) do
			handler[func](handler, self, event, unit, ...)
		end
	end
end

-- Do a full update OnShow, and stop watching for events when it's not visible
local function OnShow(self)
	-- Reset the event handler
	self:SetScript("OnEvent", OnEvent)
	Units:CheckUnitGUID(self)
end

local function OnHide(self)
	self:SetScript("OnEvent", nil)
	
	-- If it's a volatile such as target or focus, next time it's shown it has to do an update
	if( self.unitType ~= "party" and self.unitType ~= "partypet" and self.unitType ~= "partytarget" and self.unitType ~= "player" and self.unitType ~= "raid" ) then
		self.unitGUID = nil
	end
end

-- For targettarget/focustarget/etc units that don't give us real events
local function TargetUnitUpdate(self, elapsed)
	self.timeElapsed = self.timeElapsed + elapsed
	
	if( self.timeElapsed >= 0.50 ) then
		self.timeElapsed = self.timeElapsed - 0.50
		
		-- Have to make sure the unit exists or else the frame will flash offline for a second until it hides
		if( UnitExists(self.unit) ) then
			self:FullUpdate()
		end
	end
end

-- Deal with enabling modules inside a zone
local function SetVisibility(self)
	local layoutUpdate
	local zone = select(2, IsInInstance())

	-- Selectively disable modules
	for _, module in pairs(ShadowUF.moduleOrder) do
		if( module.OnEnable and module.OnDisable and ShadowUF.db.profile.units[self.unitType][module.moduleKey] ) then
			local key = module.moduleKey
			local enabled = ShadowUF.db.profile.units[self.unitType][key].enabled
			
			-- These modules have mini-modules, the entire module should be enabled if at least one is enabled, and disabled if all are disabled
			if( key == "auras" or key == "indicators" or key == "highlight" ) then
				enabled = nil
				for _, option in pairs(ShadowUF.db.profile.units[self.unitType][key]) do
					if( type(option) == "table" and option.enabled or option == true ) then
						enabled = true
						break
					end
				end
			end
			
			-- In an actual zone, check to see if we have an override for the zone
			if( zone ~= "none" ) then
				if( ShadowUF.db.profile.visibility[zone][self.unitType .. key] == false ) then
					enabled = nil
				elseif( ShadowUF.db.profile.visibility[zone][self.unitType .. key] == true ) then
					enabled = true
				end
			end
			
			-- Force disable modules for people who aren't the class for it
			if( module.moduleClass and module.moduleClass ~= playerClass ) then
				enabled = nil
			end
						
			-- Module isn't enabled all the time, only in this zone so we need to force it to be enabled
			if( not self.visibility[key] and enabled ) then
				module:OnEnable(self)
				layoutUpdate = true
			elseif( self.visibility[key] and not enabled ) then
				module:OnDisable(self)
				layoutUpdate = true
			end
			
			self.visibility[key] = enabled or nil
		end
	end
	
	-- We had a module update, force a full layout update of this frame
	if( layoutUpdate ) then
		ShadowUF.Layout:Load(self)
	end
end

-- Vehicles do not always return their data right away, a pure OnUpdate check seems to be the most accurate unfortunately
local function checkVehicleData(self, elapsed)
	self.timeElapsed = self.timeElapsed + elapsed
	if( self.timeElapsed >= 0.50 ) then
		self.timeElapsed = 0
		self.dataAttempts = self.dataAttempts + 1
		
		-- Took too long to get data, revert the frame to not being inside a vehicle
		if( self.dataAttempts >= 6 ) then
			self.timeElapsed = nil
			self.dataAttempts = nil
			self:SetScript("OnUpdate", nil)

			self.inVehicle = false
			self.unit = frame.unitOwner
			self:FullUpdate()
			
		-- Got data, stop checking and do a full frame update
		elseif( UnitIsConnected(self.unit) or UnitHealthMax(self.unit) > 0 ) then
			self.timeElapsed = nil
			self.dataAttempts = nil
			self:SetScript("OnUpdate", nil)
			
			self.unitGUID = UnitGUID(self.unit)
			self:FullUpdate()
		end
	end
end 

-- Check if a unit entered a vehicle
function Units:CheckVehicleStatus(frame, event, unit)
	if( event and frame.unitOwner ~= unit ) then return end
	
	-- Not in a vehicle yet, and they entered one that has a UI 
	if( not frame.inVehicle and UnitHasVehicleUI(frame.unitOwner) and not ShadowUF.db.profile.units[frame.unitType].disableVehicle ) then
		frame.inVehicle = true
		frame.unit = frame.vehicleUnit

		if( not UnitIsConnected(frame.unit) or UnitHealthMax(frame.unit) == 0 ) then
			frame.timeElapsed = 0
			frame.dataAttempts = 0
			frame:SetScript("OnUpdate", checkVehicleData)
		else
			frame.unitGUID = UnitGUID(frame.unit)
			frame:FullUpdate()
		end
		
		-- Keep track of what the players current unit is supposed to be, so things like auras can figure it out
		if( frame.unitOwner == "player" ) then
			ShadowUF.playerUnit = frame.unit
		end
				
	-- Was in a vehicle, no longer has a UI
	elseif( frame.inVehicle and ( not UnitHasVehicleUI(frame.unitOwner) or ShadowUF.db.profile.units[frame.unitType].disableVehicle ) ) then
		frame.inVehicle = false
		frame.unit = frame.unitOwner
		frame.unitGUID = UnitGUID(frame.unit)
		frame:FullUpdate()

		if( frame.unitOwner == "player" ) then
			ShadowUF.playerUnit = frame.unitOwner
		end
	end
end

-- The argument from UNIT_PET is the pets owner, so the player summoning a new pet gets "player", party1 summoning a new pet gets "party1" and so on
function Units:CheckUnitUpdated(frame, event, unit)
	if( unit == frame.unitRealOwner and UnitExists(frame.unit) ) then
		frame:FullUpdate()
	end
end

-- Handles checking for GUID changes for doing a full update, this fixes frames sometimes showing the wrong unit when they change
function Units:CheckUnitGUID(frame)
	local guid = frame.unit and UnitGUID(frame.unit)
	if( guid and guid ~= frame.unitGUID ) then
		frame:FullUpdate()
	end
	
	frame.unitGUID = guid
end

function Units:CheckUnitGUIDOrStatus(frame)
	-- It's possible for threat (and healcomm at that) to bug out if they are active on a unit who does
	-- this makes sure that we do a full update when they die or were offline and come back regardless of what UnitGUID says
	if( ( frame.isDead and not UnitIsDeadOrGhost(frame.unit) ) or ( frame.isOffline and UnitIsConnected(frame.unit) ) ) then
		frame:FullUpdate()
		return

	-- Alright, so heres why this is done
	-- When raid1, raid2, raid3 are in a group with each other and raid1 or raid2 are in a vehicle and get kicked
	-- OnAttributeChanged won't do anything because the frame is already setup, however the issue will have the active unit
	-- be non-existance while the primary unit is. So if we see they're in a vehicle with this case, we force the full update to get the vehicle change
	elseif( frame.inVehicle and not UnitExists(frame.unit) and UnitExists(frame.unitOwner) ) then
		self:CheckVehicleStatus(frame)
	end

	local guid = frame.unit and UnitGUID(frame.unit)
	if( guid and guid ~= frame.unitGUID ) then
		frame:FullUpdate()
	end
	
	frame.unitGUID = guid
end

local function ShowMenu(self)
	local menuFrame
	if( self.unit == "player" ) then
		menuFrame = PlayerFrameDropDown
	elseif( self.unitRealType == "party" ) then
		menuFrame = getglobal("PartyMemberFrame" .. self.unitID .. "DropDown")
	elseif( self.unitRealType == "raid" ) then
		menuFrame = FriendsDropDown
		menuFrame.displayMode = "MENU"
		menuFrame.initialize = RaidFrameDropDown_Initialize
		menuFrame.userData = self.unitID
	else
		return
	end	
	
	self.dropdownMenu = menuFrame
	
	HideDropDownMenu(1)
	menuFrame.unit = self.unitOwner
	menuFrame.name = UnitName(self.unitOwner)
	menuFrame.id = self.unitID
	ToggleDropDownMenu(1, nil, menuFrame, "cursor")
end

-- More fun with sorting, due to sorting magic we have to check if we want to create stuff when the frame changes of partys too
local function createChildUnits(self)
	if( self.unitType ~= "party" or self.unitRealType ~= "party" ) then return end
	if( loadedUnits.partypet ) then
		Units:LoadChildUnit(unitFrames.party, self.unit, "partypet", "partypet" .. self.unitID)
	end
	
	if( loadedUnits.partytarget ) then
		Units:LoadChildUnit(unitFrames.party, self.unit, "partytarget", "party" .. self.unitID .. "target")
	end
end

-- Attribute set, something changed
-- unit = Active unitid
-- unitID = Just the number from the unitid
-- unitType = Unitid minus numbers in it, used for configuration
-- unitRealType = The actual unit type, if party is shown in raid this will be "party" while unitType is still "raid"
-- unitOwner = Always the units owner even when unit changes due to vehicles
-- vehicleUnit = Unit to use when the unitOwner is in a vehicle
local function OnAttributeChanged(self, name, unit)
	-- You're the issue
	if( name ~= "unit" or not unit or unit == self.unitOwner ) then return end
	
	-- Setup identification data
	self.unit = unit
	self.unitID = tonumber(string.match(unit, "([0-9]+)"))
	self.unitOwner = unit

	-- Store it for later
	unitFrames[self.unit] = self
	
	-- Unit already exists but unitid changed, update the info we got on them
	-- Don't need to recheck the unitType and force a full update, because a raid frame can never become
	-- a party frame, or a player frame and so on
	if( self.unitRealType ) then
		self.unitRealType = string.gsub(unit, "([0-9]+)", "")
		self.vehicleUnit = self.unitOwner == "player" and "vehicle" or self.unitRealType == "party" and "partypet" .. self.unitID or self.unitRealType == "raid" and "raidpet" .. self.unitID or nil
		self.unit = self.unitOwner
		self.inVehicle = false
		createChildUnits(self)

		Units:CheckVehicleStatus(self)
		Units:CheckUnitGUID(self)
		return
	end
	
	self.unitRealType = string.gsub(unit, "([0-9]+)", "")
	self.unitType = self.unitType or self.unitRealType
	self.vehicleUnit = self.unitOwner == "player" and "vehicle" or self.unitRealType == "party" and "partypet" .. self.unitID or self.unitRealType == "raid" and "raidpet" .. self.unitID or nil

	-- Add to Clique
	ClickCastFrames = ClickCastFrames or {}
	ClickCastFrames[self] = true
	
	-- Handles switching the internal unit variable to that of their vehicle
	if( self.unit == "player" or self.unitRealType == "party" or self.unitRealType == "raid" ) then
		self:RegisterNormalEvent("UNIT_ENTERED_VEHICLE", Units, "CheckVehicleStatus")
		self:RegisterNormalEvent("UNIT_EXITED_VEHICLE", Units, "CheckVehicleStatus")
		self:RegisterUpdateFunc(Units, "CheckVehicleStatus")
	end	

	-- Pet changed, going from pet -> vehicle for one
	if( self.unit == "pet" or self.unitType == "partypet" ) then
		self.unitRealOwner = self.unit == "pet" and "player" or ShadowUF.partyUnits[self.unitID]
		self:RegisterNormalEvent("UNIT_PET", Units, "CheckUnitUpdated")
		
		if( self.unit == "pet" ) then
			self.dropdownMenu = PetFrameDropDown
			self:SetAttribute("_menu", PetFrame.menu)
			self:SetAttribute("disableVehicleSwap", ShadowUF.db.profile.units.player.disableVehicle)
		else
			self:SetAttribute("disableVehicleSwap", ShadowUF.db.profile.units.party.disableVehicle)
		end
	
		-- Logged out in a vehicle
		if( UnitHasVehicleUI(self.unitRealOwner) ) then
			self:SetAttribute("unitIsVehicle", true)
		end

		-- Hide any pet that became a vehicle, we detect this by the owner being untargetable but they have a pet out
		stateMonitor:WrapScript(self, "OnAttributeChanged", [[
			if( name == "state-vehicleupdated" ) then
				self:SetAttribute("unitIsVehicle", value == "vehicle" and true or false)
			elseif( name == "state-unitexists" ) then
				if( not value or ( not self:GetAttribute("disableVehicleSwap") and self:GetAttribute("unitIsVehicle") ) ) then
					self:Hide()
				elseif( value ) then
					self:Show()
				end
			elseif( name == "disablevehicleswap" ) then
				if( value and self:GetAttribute("state-unitexists") ) then
					self:Show()
				elseif( not self:GetAttribute("state-unitexists") or ( not value and self:GetAttribute("unitIsVehicle") ) ) then 
					self:Hide()
				end
			end
		]])
		RegisterStateDriver(self, "vehicleupdated", string.format("[target=%s, nohelp, noharm] vehicle; [target=%s, exists] pet", self.unitRealOwner, self.unit))

	-- Automatically do a full update on target change
	elseif( self.unit == "target" ) then
		self.dropdownMenu = TargetFrameDropDown
		self:SetAttribute("_menu", TargetFrame.menu)
		self:RegisterNormalEvent("PLAYER_TARGET_CHANGED", Units, "CheckUnitGUID")

	-- Automatically do a full update on focus change
	elseif( self.unit == "focus" ) then
		self.dropdownMenu = FocusFrameDropDown
		self:SetAttribute("_menu", FocusFrame.menu)
		self:RegisterNormalEvent("PLAYER_FOCUS_CHANGED", Units, "CheckUnitGUID")
				
	elseif( self.unit == "player" ) then
		self:SetAttribute("toggleForVehicle", true)
		
		-- Force a full update when the player is alive to prevent freezes when releasing in a zone that forces a ressurect (naxx/tk/etc)
		self:RegisterNormalEvent("PLAYER_ALIVE", self, "FullUpdate")
	
	-- Check for a unit guid to do a full update
	elseif( self.unitRealType == "raid" ) then
		self:RegisterNormalEvent("RAID_ROSTER_UPDATE", Units, "CheckUnitGUIDOrStatus")
		
	-- Party members need to watch for changes
	elseif( self.unitRealType == "party" ) then
		self:RegisterNormalEvent("PARTY_MEMBERS_CHANGED", Units, "CheckUnitGUIDOrStatus")
		createChildUnits(self)

	-- *target units are not real units, thus they do not receive events and must be polled for data
	elseif( string.match(self.unit, "%w+target") ) then
		self.timeElapsed = 0
		self:SetScript("OnUpdate", TargetUnitUpdate)
		
		-- Speeds up updating units when their owner changes target, if party1 changes target then party1target is force updated, if target changes target
		-- then targettarget and targettargettarget are also force updated
		if( self.unitType == "partytarget" ) then
			self.unitRealOwner = ShadowUF.partyUnits[self.unitID]
		
		elseif( self.unit == "focustarget" ) then
			self.unitRealOwner = "focus"
			self:RegisterNormalEvent("PLAYER_FOCUS_CHANGED", Units, "CheckUnitGUID")
		elseif( self.unit == "targettarget" or self.unit == "targettargettarget" ) then
			self.unitRealOwner = "target"
			self:RegisterNormalEvent("PLAYER_TARGET_CHANGED", Units, "CheckUnitGUID")
		end

		self:RegisterNormalEvent("UNIT_TARGET", Units, "CheckUnitUpdated")
	end
	
	-- No secure menu set, default to tainted
	if( not self:GetAttribute("_menu") ) then
		self.menu = ShowMenu
	end
		
	-- Update module status
	self:SetVisibility()
	
	-- Check for any unit changes
	Units:CheckUnitGUID(self)
end

-- Both of these internally call _blah and since you can't modify attributes starting with _ while in combat (period event with a secure call)
-- we have to queue this to run when combat drops, fortunately it doesn't matter. Pets won't work if created in combat anyway the first time!
local function wrapSecureCode(self)
	self:SetFrameRef("partyChildren", unitFrames.party.childContainer)
	self:WrapScript(self, "OnAttributeChanged", [[
		if( name ~= "forceposition" and ( name ~= "unit" or not value or value == self:GetAttribute("lastUnitSet") ) ) then return end
		
		-- Unless an update is forced, this will prevent it from repositioning every time the unit "changes"
		if( name == "unit" ) then
			self:SetAttribute("lastUnitSet", value)
		end
		
		-- This creates a new table when the unit completely changes, per changed unit frame for parties meaning 5 tables worst case
		-- I'll clean this up later though
		local children = table.new(self:GetFrameRef("partyChildren"):GetChildren())
		for _, child in pairs(children) do
			if( child:GetAttribute("parentUnit") == self:GetAttribute("unit") ) then
				child:ClearAllPoints()
				child:SetPoint(child:GetAttribute("framePoint"), self, child:GetAttribute("frameRelative"), child:GetAttribute("frameX"), child:GetAttribute("frameY"))
			end
		end
	]])
end

-- Header unit initialized
local function initializeUnit(self)
	local unitType = self:GetParent().unitType
	local config = ShadowUF.db.profile.units[unitType]

	self.ignoreAnchor = true
	self.unitType = unitType
	self:SetAttribute("initial-height", config.height)
	self:SetAttribute("initial-width", config.width)
	self:SetAttribute("initial-scale", config.scale)
	self:SetAttribute("toggleForVehicle", true)
	
	Units:CreateUnit(self)

	-- So more fun, when the state header notices a party change it's possible the party frames itself do not have their units yet
	-- which means, when the unit changes in the party frame we need to check what frames want us as their parent and anchor them to this
	-- this should (in general) be more reliable than my previous method too
	if( unitType == "party" ) then
		if( not InCombatLockdown() ) then
			wrapSecureCode(self)
		else
			queuedCombat[self] = true
		end
	end
end

-- Show tooltip
local function OnEnter(self)
	if( not ShadowUF.db.profile.tooltipCombat or not InCombatLockdown() ) then
		UnitFrame_OnEnter(self)
	end
end

-- Reset the fact that we clamped the dropdown to the screen to be safe
DropDownList1:HookScript("OnHide", function(self)
	self:SetClampedToScreen(false)
end)

-- Reposition the dropdown
local function PostClick(self)
	if( UIDROPDOWNMENU_OPEN_MENU == self.dropdownMenu and DropDownList1:IsShown() )	 then
		DropDownList1:ClearAllPoints()
		DropDownList1:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 0, 0)
		DropDownList1:SetClampedToScreen(true)
	end
end

-- Create the generic things that we want in every secure frame regardless if it's a button or a header
function Units:CreateUnit(frame)
	frame.fullUpdates = {}
	frame.registeredEvents = {}
	frame.visibility = {}
	frame.RegisterNormalEvent = RegisterNormalEvent
	frame.RegisterUnitEvent = RegisterUnitEvent
	frame.RegisterUpdateFunc = RegisterUpdateFunc
	frame.UnregisterAll = UnregisterAll
	frame.UnregisterSingleEvent = UnregisterEvent
	frame.FullUpdate = FullUpdate
	frame.SetVisibility = SetVisibility
	frame.topFrameLevel = 5
	
	-- Ensures that text is the absolute highest thing there is
	frame.highFrame = CreateFrame("Frame", nil, frame)
	frame.highFrame:SetFrameLevel(frame.topFrameLevel + 1)
	frame.highFrame:SetAllPoints(frame)
	
	frame:SetScript("OnAttributeChanged", OnAttributeChanged)
	frame:SetScript("OnEvent", OnEvent)
	frame:SetScript("OnEnter", OnEnter)
	frame:SetScript("OnLeave", UnitFrame_OnLeave)
	frame:SetScript("OnShow", OnShow)
	frame:SetScript("OnHide", OnHide)
	frame:SetScript("PostClick", PostClick)

	frame:RegisterForClicks("AnyUp")	
	frame:SetAttribute("*type1", "target")
	frame:SetAttribute("*type2", "menu")
end

-- Update the main header
function Units:ReloadHeader(type)
	local frame = unitFrames[type]
	if( frame ) then
		self:SetFrameAttributes(frame, type)
		ShadowUF.Layout:AnchorFrame(UIParent, frame, ShadowUF.db.profile.positions[type])
	end
end

function Units:SetFrameAttributes(frame, type)
	local config = ShadowUF.db.profile.units[type]
	if( not config ) then
		return
	end
	
	local xMod = config.attribPoint == "LEFT" and 1 or config.attribPoint == "RIGHT" and -1 or 0
	local yMod = config.attribPoint == "TOP" and -1 or config.attribPoint == "BOTTOM" and 1 or 0
	
	frame:SetAttribute("point", config.attribPoint)
	frame:SetAttribute("sortMethod", config.sortMethod)
	frame:SetAttribute("sortDir", config.sortOrder)
	
	frame:SetAttribute("xOffset", config.offset * xMod)
	frame:SetAttribute("yOffset", config.offset * yMod)
			
	if( type == "raid" ) then
		local filter
		for id, enabled in pairs(config.filters) do
			if( enabled ) then
				if( filter ) then
					filter = filter .. "," .. id
				else
					filter = id
				end
			end
		end
	
		frame:SetAttribute("showRaid", true)
		frame:SetAttribute("maxColumns", config.maxColumns)
		frame:SetAttribute("unitsPerColumn", config.unitsPerColumn)
		frame:SetAttribute("columnSpacing", config.columnSpacing)
		frame:SetAttribute("columnAnchorPoint", config.attribAnchorPoint)
		frame:SetAttribute("groupFilter", filter or "1,2,3,4,5,6,7,8")
		
		if( config.groupBy == "CLASS" ) then
			frame:SetAttribute("groupingOrder", "DEATHKNIGHT,DRUID,HUNTER,MAGE,PALADIN,PRIEST,ROGUE,SHAMAN,WARLOCK,WARRIOR")
			frame:SetAttribute("groupBy", "CLASS")
		else
			frame:SetAttribute("groupingOrder", "1,2,3,4,5,6,7,8")
			frame:SetAttribute("groupBy", "GROUP")
		end
		
	-- Update party frames to not show anyone if they should be in raids
	elseif( type == "party" ) then
		frame:SetAttribute("showParty", ( not config.showAsRaid or not ShadowUF.db.profile.units.raid.enabled ) and true or false)
		frame:SetAttribute("showPlayer", config.showPlayer)
	end

	-- Update the raid frames to if they should be showing raid or party
	if( unitFrames.raid and unitFrames.party ) then
		unitFrames.raid:SetAttribute("showParty", not unitFrames.party:GetAttribute("showParty"))
		unitFrames.raid:SetAttribute("showPlayer", unitFrames.party:GetAttribute("showPlayer"))
	end
	
	-- Need to update our flags on the state monitor so it knows what to do
	stateMonitor:SetAttribute("hideSemiRaid", ShadowUF.db.profile.units.party.hideSemiRaid)
	stateMonitor:SetAttribute("hideAnyRaid", ShadowUF.db.profile.units.party.hideAnyRaid)
end

-- Load a single unit such as player, target, pet, etc
function Units:LoadUnit(config, unit)
	-- Already be loaded, just enable
	if( unitFrames[unit] ) then
		RegisterUnitWatch(unitFrames[unit], unit == "pet")
		return
	end
	
	local frame = CreateFrame("Button", "SUFUnit" .. unit, UIParent, "SecureUnitButtonTemplate")
	self:CreateUnit(frame)
	frame:SetAttribute("unit", unit)

	unitFrames[unit] = frame
		
	-- Annd lets get this going
	RegisterUnitWatch(frame, unit == "pet")
end

-- Load a header unit, party or raid
function Units:LoadGroupHeader(config, type)
	if( unitFrames[type] ) then
		unitFrames[type]:Show()
		
		if( type == "party" ) then
			stateMonitor:SetAttribute("partyDisabled", nil)
		end
		return
	end
	
	local headerFrame = CreateFrame("Frame", "SUFHeader" .. type, UIParent, "SecureGroupHeaderTemplate")
	unitFrames[type] = headerFrame

	self:SetFrameAttributes(headerFrame, type)
	
	if( type == "party" ) then
		headerFrame:SetAttribute("template", "SecureUnitButtonTemplate,SecureHandlerBaseTemplate")
	else
		headerFrame:SetAttribute("template", "SecureUnitButtonTemplate")
	end
	
	headerFrame:SetAttribute("initial-unitWatch", true)
	headerFrame.initialConfigFunction = initializeUnit
	headerFrame.isHeaderFrame = true
	headerFrame.unitType = type
	headerFrame:UnregisterEvent("UNIT_NAME_UPDATE")
	
	ShadowUF.Layout:AnchorFrame(UIParent, headerFrame, ShadowUF.db.profile.positions[type])
	
	-- We have to do party hiding based off raid as a state driver so that we can smoothly hide the party frames based off of combat and such
	-- technically this isn't the cleanest solution because party frames will still have unit watches active
	-- but this isn't as big of a deal, because SUF automatically will unregister the OnEvent for party frames while hidden
	-- will add extra code to fully disable them later on, but the now this is a good enough solution.
	if( type == "party" ) then
		headerFrame.childContainer = CreateFrame("Frame", nil, headerFrame,"SecureHandlerBaseTemplate")
		stateMonitor:SetFrameRef("partyHeader", headerFrame)
		stateMonitor:WrapScript(stateMonitor, "OnAttributeChanged", [[
			if( name ~= "state-raidmonitor" and name ~= "partydisabled" and name ~= "hideanyraid" and name ~= "hidesemiraid" ) then return end
			if( self:GetAttribute("partyDisabled") ) then return end
			
			if( self:GetAttribute("hideAnyRaid") and ( self:GetAttribute("state-raidmonitor") == "raid1" or self:GetAttribute("state-raidmonitor") == "raid5" ) ) then
				self:GetFrameRef("partyHeader"):Hide()
			elseif( self:GetAttribute("hideSemiRaid") and self:GetAttribute("state-raidmonitor") == "raid5" ) then
				self:GetFrameRef("partyHeader"):Hide()
			else
				self:GetFrameRef("partyHeader"):Show()
			end
		]])
		RegisterStateDriver(stateMonitor, "raidmonitor", "[target=raid5, exists] raid5; [target=raid1, exists] raid1; none")
	else
		headerFrame:Show()
	end
end

function Units:SetChildFrameAttributes(frame, type)
	-- If it's not a party pet or party target then it can be manually positioned
	if( type ~= "partypet" and type ~= "partytarget" ) then
		local parentFrame = unitFrames[frame:GetAttribute("parentUnit")]
		if( parentFrame ) then
			ShadowUF.Layout:AnchorFrame(parentFrame, frame, ShadowUF.db.profile.positions[type])
		end
	else
		-- Save the positioning information so the secure stuff can access it
		frame:SetAttribute("framePoint", ShadowUF.Layout:GetPoint(ShadowUF.db.profile.positions[type].anchorPoint))
		frame:SetAttribute("frameRelative", ShadowUF.Layout:GetRelative(ShadowUF.db.profile.positions[type].anchorPoint))
		frame:SetAttribute("frameX", ShadowUF.db.profile.positions[type].x)
		frame:SetAttribute("frameY", ShadowUF.db.profile.positions[type].y)
		
		-- If this units parent frame exists, force it to update
		local parentFrame = unitFrames[frame:GetAttribute("parentUnit")]
		if( parentFrame ) then
			parentFrame:SetAttribute("forceposition", GetTime())
		end
	end
end

-- Load a unit that is a child of another unit (party pet/party target)
function Units:LoadChildUnit(header, parentUnit, type, unit)
	if( InCombatLockdown() ) then
		queuedCombat[unit] = unitFrames[parentUnit]
		return
	elseif( unitFrames[unit] ) then
		RegisterUnitWatch(unitFrames[unit], type == "partypet")
		self:SetChildFrameAttributes(unitFrames[unit], type)
		return
	end
		
	-- Now we can create the actual frame
	local frame = CreateFrame("Button", "SUFUnit" .. unit, header.childContainer or header, "SecureUnitButtonTemplate")
	self:CreateUnit(frame)

	frame.isChildUnit = true
	frame.ignoreAnchor = true
	frame.header = header
	frame:SetFrameStrata("LOW")
	frame:SetAttribute("unit", unit)
	frame:SetAttribute("parentUnit", parentUnit)
	
	unitFrames[unit] = frame

	RegisterUnitWatch(frame, type == "partypet")
		
	-- Setup the attributes to watch
	self:SetChildFrameAttributes(frame, type)
end

-- Initialize units
function Units:InitializeFrame(config, type)
	if( loadedUnits[type] ) then return end
	loadedUnits[type] = true
	
	if( type == "party" or type == "raid" ) then
		self:LoadGroupHeader(config, type)
	-- Party pet/target are not created until their owners are, this will only initialize them if the owners were loaded already
	elseif( type == "partypet" ) then
		for id, unit in pairs(ShadowUF.partyUnits) do
			if( unitFrames[unit] ) then
				Units:LoadChildUnit(unitFrames.party, unit, type, "partypet" .. id)
			end
		end
	elseif( type == "partytarget" ) then
		for id, unit in pairs(ShadowUF.partyUnits) do
			if( unitFrames[unit] ) then
				Units:LoadChildUnit(unitFrames.party, unit, type, "party" .. id .. "target")
			end
		end
	else
		self:LoadUnit(config, type)
	end
end

-- Uninitialize units
function Units:UninitializeFrame(config, type)
	if( not loadedUnits[type] ) then return end
	if( type == "party" ) then
		stateMonitor:SetAttribute("partyDisabled", true)
	end
	
	if( type == "raid" and loadedUnits.party ) then
		unitFrames.party:SetAttribute("showParty", true)
	end

	loadedUnits[type] = nil
		
	-- Disable all frames of this time
	for _, frame in pairs(unitFrames) do
		if( frame.unitType == type ) then
			UnregisterUnitWatch(frame)
			frame:Hide()
		end
	end
end

-- Profile changed, reload units
function Units:ProfileChanged()
	-- Reset the anchors for all frames to prevent X is dependant on Y
	for _, frame in pairs(unitFrames) do
		if( frame.unit ) then
			frame:ClearAllPoints()
		end
	end
	
	-- Force all of the module changes
	for _, frame in pairs(unitFrames) do
		if( frame.unit and loadedUnits[frame.unitType] ) then
			-- Force all enabled modules to disable
			for key, module in pairs(ShadowUF.modules) do
				if( frame[key] and frame.visibility[key] ) then
					frame.visibility[key] = nil
					module:OnDisable(frame)
				end
			end
			
			-- Now enable whatever we need to
			frame:SetVisibility()
			ShadowUF.Layout:Load(frame)
			frame:FullUpdate()
			
			-- Update the childs positioning too
			if( frame.isChildUnit ) then
				ShadowUF.Units:SetChildFrameAttributes(frame, frame.unitType)
			end
		end
	end
	
	-- Force headers to update
	self:ReloadHeader("raid")
	self:ReloadHeader("party")
end

-- Small helper function for creating bars with
function Units:CreateBar(parent)
	local bar = CreateFrame("StatusBar", nil, parent)
	bar:SetFrameLevel(parent.topFrameLevel)
	bar.parent = parent
	
	bar.background = bar:CreateTexture(nil, "BORDER")
	bar.background:SetHeight(1)
	bar.background:SetWidth(1)
	bar.background:SetAllPoints(bar)

	return bar
end

-- Deal with zone changes for enabling modules
local instanceType
function Units:CheckPlayerZone(force)
	local instance = select(2, IsInInstance())
	if( instance == instanceType and not force ) then return end
	instanceType = instance
	
	ShadowUF:LoadUnits()
	for _, frame in pairs(unitFrames) do
		if( frame.unit ) then
			frame:SetVisibility()
			
			-- Auras are enabled so will need to check if the filter has to change
			if( frame.visibility.auras ) then
				ShadowUF.modules.auras:UpdateFilter(frame)
			end
			
			if( UnitExists(frame.unit) ) then
				frame:FullUpdate()
			end
		end
	end
end

-- Handles events related to all units and not a specific one
local centralFrame = CreateFrame("Frame")
centralFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
centralFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA")
centralFrame:SetScript("OnEvent", function(self, event, unit)
	-- Check if the player changed zone types and we need to change module status, while they are dead
	-- we won't change their zone type as releasing from an instance will change the zone type without them
	-- really having left the zone
	if( event == "ZONE_CHANGED_NEW_AREA" ) then
		if( UnitIsDeadOrGhost("player") ) then
			self:RegisterEvent("PLAYER_UNGHOST")
		else
			self:UnregisterEvent("PLAYER_UNGHOST")
			Units:CheckPlayerZone()
		end
		
	-- They're alive again so they "officially" changed zone types now
	elseif( event == "PLAYER_UNGHOST" ) then
		Units:CheckPlayerZone()
		
	-- This is slightly hackish, but it suits the purpose just fine for somthing thats rarely called.
	elseif( event == "PLAYER_REGEN_ENABLED" ) then
		-- Need to check if we have to wrap it first
		for frame, needsWrap in pairs(queuedCombat) do
			if( needsWrap == true ) then
				queuedCombat[frame] = nil
				wrapSecureCode(frame)
			end
		end
		
		-- Now do all of the creation for child wrapping
		for unitID, parent in pairs(queuedCombat) do
			queuedCombat[unitID] = nil
			Units:LoadChildUnit(unitFrames[parent.unitType], parent.unit, string.gsub(unitID, "(%d+)", ""), unitID)
		end
	end
end)